@@ -19,7 +19,7 @@
 # DEALINGS IN THE SOFTWARE.
 
 from hy.models import HyObject
-from hy.util import str_type
+from hy._compat import str_type
 
 
 class HyString(HyObject, str_type):
